/*
 * code to read and set the ID3/ID3X tags on an MP3 file
 *
 * This source code is provided to you with NO WARRANTY and NO LICENSE.
 * You may do with it as you please, and I cannot be held responsible for
 * any outcome of the use of this code.  It is hereby released into the
 * public domain.
 *      - Robey Pointer
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "idtag.h"

/*
#undef ID3_STANDALONE
#undef ID3_DEBUG
*/


/* remove the trailing spaces from a string */
static void strunpad(char *str)
{
    int i = strlen(str);
   
    while ((i > 0) && (str[i-1] == ' '))
	i--;
    str[i] = 0;
}

/* copy, then pad the buffer with spaces */
static void strpad(char *dest, char *src, int len)
{
    memset(dest, ' ', len);
    if (! src)
	return;
    strncpy(dest, src, len > (signed)strlen(src) ? strlen(src) : len);
}

/*
 * read an id3 tag from the end of an mp3 file.
 * if an id3x tag is present, that will be silently read also.  (id3x
 *     extends the length of several of the id3 fields.)
 * returns 1 on success, 0 on failure
 */
int id3_read(int fd, ID3Tag *tag)
{
    char buffer[256];
    int n;

    if (lseek(fd, -128, SEEK_END) == (off_t)-1)
	return 0;
    n = read(fd, buffer, 128);
    if (n < 128)
	return 0;

    /* is this an id3 tag? */
    if (strncmp(buffer, "TAG", 3) != 0)
	return 0;	/* nope. */
    strncpy(tag->title, buffer+3, 30);
    tag->title[30] = 0;
    strncpy(tag->artist, buffer+33, 30);
    tag->artist[30] = 0;
    strncpy(tag->album, buffer+63, 30);
    tag->album[30] = 0;
    strncpy(tag->year, buffer+93, 4);
    tag->year[4] = 0;
    strncpy(tag->comment, buffer+97, 30);
    tag->comment[30] = 0;
    tag->unknown = buffer[127];
#ifdef ID3_DEBUG
    printf("* found id3 tag.\n");
#endif

    /* check for id3x tag */
    if ((lseek(fd, -384, SEEK_END) != (off_t)-1) &&
	(read(fd, buffer, 256) == 256) &&
	(strncmp(buffer, "TXG", 3) == 0)) {
	/* ok, we have an id3x tag -- fill in the extended fields */
	strncpy(tag->title+30, buffer+3, 90);
	tag->title[120] = 0;
	strncpy(tag->artist+30, buffer+93, 50);
	tag->artist[80] = 0;
	strncpy(tag->album+30, buffer+143, 50);
	tag->album[80] = 0;
	strncpy(tag->comment+30, buffer+193, 50);
	tag->comment[80] = 0;
#ifdef ID3_DEBUG
	printf("* found id3x tag.\n");
#endif
    }

    strunpad(tag->title);
    strunpad(tag->artist);
    strunpad(tag->album);
    strunpad(tag->comment);
    return 1;
}

/*
 * write an id3 tag to the end of an mp3 file.
 * if any of the tag fields are long enough to require an id3x tag also,
 *     the id3x tag is silently written too.
 * if 'write_only' is set, i won't check for existing tags -- i'll just
 *     smack new tags on the end of the file.  this is useful if you just
 *     created this mp3 file yourself (and only have the file open for
 *     writing).  if you use 'write_only' on an mp3 that already has an
 *     id3 tag, i'll be appending new ones and the old tags will occupy
 *     dead space, which is wasteful, but probably mostly harmless.
 * returns 1 on success, 0 on failure
 */
int id3_write(int fd, ID3Tag *tag, int write_only)
{
    char buffer[256];
    int has_id3 = 0, has_id3x = 0, need_id3x = 0;

    if (! write_only) {
	/* look for existing tags */
	if ((lseek(fd, -384, SEEK_END) != (off_t)-1) &&
	    (read(fd, buffer, 3) == 3) &&
	    (strncmp(buffer, "TXG", 3) == 0)) {
#ifdef ID3_DEBUG
    printf("* found id3x tag.\n");
#endif
	    has_id3x = 1;
	}
	if ((lseek(fd, -128, SEEK_END) != (off_t)-1) &&
	    (read(fd, buffer, 3) == 3) &&
	    (strncmp(buffer, "TAG", 3) == 0)) {
#ifdef ID3_DEBUG
    printf("* found id3 tag.\n");
#endif
	    has_id3 = 1;
	}
    }

    /* will we need to create an id3x tag? */
    if ((strlen(tag->title) > 30) || (strlen(tag->artist) > 30) ||
	(strlen(tag->album) > 30) || (strlen(tag->comment) > 30))
	need_id3x = 1;

    if (has_id3x || need_id3x) {
	/* write id3x tag */
	if (has_id3x) {
	    if (lseek(fd, -384, SEEK_END) == (off_t)-1)
		return 0;
	} else if (has_id3) {
	    /* have an old id3 tag, but not an id3x tag -- we need to write
	     * both.  rewind to write over the id3 tag.
	     */
	    if (lseek(fd, -128, SEEK_END) == (off_t)-1)
		return 0;
	}
	strcpy(buffer, "TXG");
	memset(buffer+3, ' ', 240);
	if (strlen(tag->title) > 30)
	    strpad(buffer+3, tag->title+30, 90);
	if (strlen(tag->artist) > 30)
	    strpad(buffer+93, tag->artist+30, 50);
	if (strlen(tag->album) > 30)
	    strpad(buffer+143, tag->album+30, 50);
	if (strlen(tag->comment) > 30)
	    strpad(buffer+193, tag->comment+30, 50);
	memset(buffer+243, 0, 13);

	if (write(fd, buffer, 256) != 256)
	    return 0;
#ifdef ID3_DEBUG
	printf("* wrote id3x tag.\n");
#endif
    }

    /* now write the "classic" id3 tag */
    if (has_id3 && !need_id3x) {
	/* if we needed to write an id3x tag, we're positioned where we
	 * should be already.
	 */
	if (lseek(fd, -128, SEEK_END) == (off_t)-1)
	    return 0;
    }
    if (!has_id3 && !has_id3x && !write_only) {
	if (lseek(fd, 0, SEEK_END) == (off_t)-1)
	    return 0;
    }

    strcpy(buffer, "TAG");
    strpad(buffer+3, tag->title, 30);
    strpad(buffer+33, tag->artist, 30);
    strpad(buffer+63, tag->album, 30);
    strpad(buffer+93, tag->year, 4);
    strpad(buffer+97, tag->comment, 30);
    buffer[127] = 0xFF;
#ifdef ID3_DEBUG
    printf("* wrote id3 tag.\n");
#endif
    return (write(fd, buffer, 128) == 128);
}


#ifdef ID3_STANDALONE

extern int optind;

static void usage(char *argv0)
{
    printf("\n");
    printf("usage: %s [options] <mp3-file>\n", argv0);
    printf("    displays the id3 tag for an mp3 file; if options are\n");
    printf("    given, writes a new id3 tag for an mp3 file\n");
    printf("\n");
    printf("options:\n");
    printf("    -t <title>      change id3 song title\n");
    printf("    -a <artist>     change id3 artist\n");
    printf("    -l <album>      change id3 album name\n");
    printf("    -y <year>       change id3 year\n");
    printf("    -c <comment>    change id3 comment\n");
    printf("\n");
}

int main(int argc, char **argv)
{
    char *title = NULL, *artist = NULL, *album = NULL, *comment = NULL,
	*year = NULL;
    ID3Tag tag;
    int c, fd;
    int writing = 0;

    while ((c = getopt(argc, argv, "t:a:l:y:c:")) != EOF) {
	switch (c) {
	case 't':
	    title = optarg;
	    break;
	case 'a':
	    artist = optarg;
	    break;
	case 'l':
	    album = optarg;
	    break;
	case 'c':
	    comment = optarg;
	    break;
	case 'y':
	    year = optarg;
	    break;
	default:
	    usage(argv[0]);
	    exit(1);
	}
    }

    if (optind >= argc) {
	printf("\nerror: mp3 filename required\n");
	usage(argv[0]);
	exit(1);
    }
    if (title || artist || album || comment || year)
	writing = 1;

    if (strcmp(argv[optind], "-") == 0) {
	fd = 0;		/* stdin */
    } else {
	fd = open(argv[optind], writing ? O_RDWR : O_RDONLY);
	if (fd < 0) {
	    printf("\nerror: can't open file '%s'\n\n", argv[optind]);
	    exit(1);
	}
    }

    if (! id3_read(fd, &tag)) {
	if (!writing) {
	    printf("\nno id3 tag\n\n");
	    exit(1);
	}
	/* clear out the idtag then */
	tag.title[0] = 0;
	tag.artist[0] = 0;
	tag.album[0] = 0;
	tag.year[0] = 0;
	tag.comment[0] = 0;
    }

    if (writing) {
	/* change fields */
	if (title)
	    strncpy(tag.title, title, 120);
	if (artist)
	    strncpy(tag.artist, artist, 80);
	if (album)
	    strncpy(tag.album, album, 80);
	if (year)
	    strncpy(tag.year, year, 4);
	if (comment)
	    strncpy(tag.comment, comment, 80);

	/* now write them */
	if (! id3_write(fd, &tag, 0)) {
	    printf("\nfailed to write new id3 tag\n\n");
	    exit(1);
	}
    }

    printf("\n");
    printf("Title:   %.120s\n", tag.title);
    printf("Artist:  %.80s\n", tag.artist);
    printf("Album:   %.80s\n", tag.album);
    printf("Year:    %.4s\n", tag.year);
    printf("Comment: %.80s\n", tag.comment);
    printf("\n");

    exit(0);
}

#endif
